#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#define	N			(6*1024*1024)

typedef unsigned int  uint;
typedef unsigned char uchar;

#define LOG2_WARP_SIZE 			5
#define WARP_SIZE 				32
#define TAG_MASK 				0x07FFFFFFU				
#define	NUM_BINS				256
#define NUM_WARPS 				6			//Warps == subhistograms per threadblock
#define MERGE_THREADBLOCK_SIZE 256

inline __device__ void addByte ( volatile uint * warpHist, uint data, uint threadTag )
{
        uint count;
		
        do
		{
            count           = warpHist [data] & TAG_MASK;
            count           = threadTag | (count + 1);
            warpHist [data] = count;
        }
		while ( warpHist [data] != count );
}

inline __device__ void addWord ( volatile uint * warpHist, uint data, uint tag )
{
    addByte ( warpHist, (data >>  0) & 0xFFU, tag );
    addByte ( warpHist, (data >>  8) & 0xFFU, tag );
    addByte ( warpHist, (data >> 16) & 0xFFU, tag );
    addByte ( warpHist, (data >> 24) & 0xFFU, tag );
	
}

__global__ void histogramKernel ( uint * partialHistograms, uint * data, uint dataCount )
{
							//     warp
    __shared__ uint hist [NUM_BINS * NUM_WARPS];
    uint          * warpHist = hist + (threadIdx.x >> LOG2_WARP_SIZE) * NUM_BINS;

							//   
	#pragma unroll
    for ( uint i = 0; i < NUM_BINS / WARP_SIZE; i++ )
       hist [threadIdx.x + i * NUM_WARPS * WARP_SIZE] = 0;

							//  id    
    uint tag = threadIdx.x << (32 - LOG2_WARP_SIZE);

    __syncthreads();
	
							//      
    for ( uint pos = blockIdx.x * blockDim.x + threadIdx.x; pos < dataCount; pos += blockDim.x * gridDim.x )
	{
        uint d = data [pos];
		
        addWord ( warpHist, d, tag );
    }

    __syncthreads();	
						//           
						// 192     256  
    for ( uint bin = threadIdx.x; bin < NUM_BINS; bin += NUM_WARPS * WARP_SIZE )
	{
        uint sum = 0;

        for ( uint i = 0; i < NUM_WARPS; i++ )
            sum += hist [bin + i * NUM_BINS] & TAG_MASK;

        partialHistograms [blockIdx.x * NUM_BINS + bin] = sum;
    }
	
}

//  ,      NUM_BINS 
__global__ void mergeHistogramKernel ( uint * outHistogram, uint * partialHistograms, uint histogramCount )
{
    uint sum = 0;
	
    for ( uint i = threadIdx.x; i < histogramCount; i += 256 )
        sum += partialHistograms [blockIdx.x + i * NUM_BINS];

    __shared__ uint data [NUM_BINS];
	
    data [threadIdx.x] = sum;

    for ( uint stride = NUM_BINS / 2; stride > 0; stride >>= 1)
	{
        __syncthreads();
		
        if ( threadIdx.x < stride )
            data [threadIdx.x] += data [threadIdx.x + stride];
    }

    if ( threadIdx.x == 0 )
        outHistogram [blockIdx.x] = data [0];
}

void histogram ( uint * histogram, void * dataDev, uint byteCount )
{
    assert( byteCount % 4 == 0 );
	
	int	   n                 = byteCount / 4;
	int	   numBlocks         = n / (NUM_WARPS * WARP_SIZE);
	int	   numPartials       = 240;
	uint * partialHistograms = NULL;
	
    cudaMalloc ( (void **) &partialHistograms, numPartials * NUM_BINS * sizeof ( uint ) );
	
    histogramKernel<<<dim3 ( numPartials ), dim3 (NUM_WARPS * WARP_SIZE) >>> ( partialHistograms, (uint *) dataDev, n );

    mergeHistogramKernel<<<dim3(NUM_BINS), dim3(256)>>> ( histogram, partialHistograms, numPartials );

    cudaFree ( partialHistograms );
}

void randomInit ( uint * a, int n, uint * h )
{
	for ( int i = 0; i < n; i++ )
	{
		uchar b1 = rand () & 0xFF;
		uchar b2 = rand () & 0xFF;
		uchar b3 = rand () & 0xFF;
		uchar b4 = rand () & 0xFF;
		
		a [i] = b1 | (b2 << 8) | (b3 << 16) | (b4 << 24);
		
		h [b1]++;
		h [b2]++;
		h [b3]++;
		h [b4]++;
	}
}

int main ( int argc, char * argv [] )
{
	uint	* a    = new uint [N];
	uint	* hDev = NULL;
	uint	* aDev = NULL;
	uint	  h     [NUM_BINS];
	uint	  hHost [NUM_BINS];
    cudaEvent_t start, stop;
    float 		gpuTime = 0.0f;
	
	memset     ( hHost, 0, sizeof ( hHost ) );
	randomInit ( a, N, hHost );

    cudaEventCreate ( &start );
    cudaEventCreate ( &stop );
	
    cudaEventRecord ( start, 0 );
	
	cudaMalloc ( (void **) &aDev, N * sizeof ( uint ) );
	cudaMalloc ( (void **) &hDev, NUM_BINS * sizeof ( uint ) );
    cudaMemcpy ( aDev, a, N * sizeof ( uint ), cudaMemcpyHostToDevice );
	
	histogram  ( hDev, aDev, 4 * N );
	
    cudaMemcpy ( h, hDev, NUM_BINS * sizeof ( uint ), cudaMemcpyDeviceToHost );
	cudaFree   ( aDev );
	cudaFree   ( hDev );

    cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	printf ( "Elapsed time: %.2f\n", gpuTime );
	
	for ( int i = 0; i < NUM_BINS; i++ )
		if ( h [i] != hHost [i] )
			printf ( "Diff at %d - %d, %d\n", i, h [i], hHost [i] );
			
	delete a;
	
	return 0;
}