#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#define	NUM_BINS	64
#define	N			(64*255*64)		//255 bytes for each thread for 1 block

typedef unsigned int  uint;
typedef unsigned char uchar;

inline __device__ void addByte ( uchar * base, uint data )
{
	base[64*data]++;
}

inline __device__ void addWord ( uchar * base, uint data )
{
				//  6     .
    addByte ( base, (data >>  2) & 0x3FU );
    addByte ( base, (data >> 10) & 0x3FU );
    addByte ( base, (data >> 18) & 0x3FU );
    addByte ( base, (data >> 26) & 0x3FU );
}

__global__ void histogram64Kernel ( uint * partialHistograms, uint * data, uint dataCount)
{
	int	tid = (threadIdx.x >> 4) | ((threadIdx.x & 0x0F) << 2);
	
				//     .
    __shared__ uchar hist [64*64];
    uchar * base = hist + tid;

				//  
    #pragma unroll
    for ( int i = 0; i < 64 / 4; i++ )
        ((uint *)hist)[threadIdx.x + i * 64] = 0;

    __syncthreads ();
	
    for ( uint pos = blockIdx.x*blockDim.x + threadIdx.x; pos < dataCount; pos += blockDim.x*gridDim.x )
	{
        uint	d = data [pos];
		
        addWord ( base, d );
    }

				//Accumulate per-thread histograms into per-block and write to global memory
    __syncthreads ();
	
	uint	sum = 0;
	uint	pos = 0;
	
	base = hist + threadIdx.x * 64;
	
	for ( int i = 0; i < 64; i++, pos += 64 )
		sum += base [i];

    partialHistograms [blockIdx.x * 64 + threadIdx.x] = sum;
}

#define MERGE_THREADBLOCK_SIZE 256

__global__ void mergeHistogram64Kernel ( uint * histogram, uint * partialHistograms, uint histogramCount )
{
    __shared__ uint data [MERGE_THREADBLOCK_SIZE];

    uint sum = 0;
	
    for ( uint i = threadIdx.x; i < histogramCount; i += MERGE_THREADBLOCK_SIZE )
        sum += partialHistograms [blockIdx.x + i * 64];
		
    data [threadIdx.x] = sum;

    for ( uint stride = MERGE_THREADBLOCK_SIZE / 2; stride > 0; stride >>= 1 )
	{
        __syncthreads ();
		
        if ( threadIdx.x < stride )
            data [threadIdx.x] += data [threadIdx.x + stride];
    }

    if ( threadIdx.x == 0 )
        histogram [blockIdx.x] = data [0];
}

void histogramm ( uint * data, uint * histogram, int byteCount )
{
		//  partialHistogram -       histogram64Kernel
	uint	sz = 64 * 255;
    uint	histogramCount = (byteCount/4 + sz - 1) / sz; //(byteCount + 254) / 255;
	uint  * partialHistograms;

    cudaMalloc ( (void **) &partialHistograms, histogramCount * 64 * sizeof ( uint ) );
	
    histogram64Kernel<<<histogramCount, 64>>> ( partialHistograms, data, byteCount / sizeof (uint) );
    mergeHistogram64Kernel<<<NUM_BINS, MERGE_THREADBLOCK_SIZE>>> ( histogram, partialHistograms, histogramCount );
	
    cudaFree ( partialHistograms );
}

void randomInit ( uint * a, int n, uint * h )
{
	for ( int i = 0; i < n; i++ )
	{
		uchar b1 = rand () & 0xFF;
		uchar b2 = rand () & 0xFF;
		uchar b3 = rand () & 0xFF;
		uchar b4 = rand () & 0xFF;
		
		a [i] = b1 | (b2 << 8) | (b3 << 16) | (b4 << 24);
		
		h [(a[i] >> 2 ) & 0x3F]++;
		h [(a[i] >> 10) & 0x3F]++;
		h [(a[i] >> 18) & 0x3F]++;
		h [(a[i] >> 26) & 0x3F]++;
	}
}

int main ( int argc, char * argv [] )
{
	uint	* a    = new uint [N];
	uint	* hDev = NULL;
	uint	* aDev = NULL;
	uint	  h     [NUM_BINS];
	uint	  hHost [NUM_BINS];
    cudaEvent_t start, stop;
    float 		gpuTime = 0.0f;
	
	memset     ( hHost, 0, sizeof ( hHost ) );
	randomInit ( a, N, hHost );

    cudaEventCreate ( &start );
    cudaEventCreate ( &stop );
	
    cudaEventRecord ( start, 0 );
	
	cudaMalloc ( (void **) &aDev, N * sizeof ( uint ) );
	cudaMalloc ( (void **) &hDev, NUM_BINS * sizeof ( uint ) );
    cudaMemcpy ( aDev, a, N * sizeof ( uint ), cudaMemcpyHostToDevice );
	
	histogramm ( aDev, hDev, N*4 );
	
    cudaMemcpy ( h, hDev, NUM_BINS * sizeof ( uint ), cudaMemcpyDeviceToHost );
	cudaFree   ( aDev );
	cudaFree   ( hDev );

    cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	printf ( "Elapsed time: %.2f\n", gpuTime );
	
	for ( int i = 0; i < NUM_BINS; i++ )
		if ( h [i] != hHost [i] )
			printf ( "Diff at %d - %d, %d\n", i, h [i], hHost [i] );
			
	delete a;
	
	return 0;
}
