//
// Perform square matrix multiplication A*At using shared memory
//

#include <stdio.h>

#define BLOCK_SIZE 	16			// submatrix size
#define	N			1024		// matrix size is N*N
#define	EPS			0.0001f

#ifndef	M_PI
	#define	M_PI	3.1415926f
#endif

__global__ void matMult ( float * a, int n, float * c )
{
	int bx = blockIdx.x;		// block index
	int by = blockIdx.y;

	int tx = threadIdx.x;		// thread index
	int ty = threadIdx.y;
	
								// Index of the first sub-matrix of A processed by the block
	int aBegin = n * BLOCK_SIZE * by;
	int aEnd = aBegin + n - 1;
								// Index of the first sub-matrix of B processed by the block
	int atBegin = n * BLOCK_SIZE * bx;
								// Step size used to iterate through the sub-matrices of B
	float sum = 0.0f;			// computed subelement
	
	for ( int ia = aBegin, iat = atBegin; ia <= aEnd; ia += BLOCK_SIZE, iat += BLOCK_SIZE )
	{
							// Shared memory for the sub-matrix of A
		__shared__ float as  [BLOCK_SIZE][BLOCK_SIZE];
							// Shared memory for the sub-matrix of At
		__shared__ float ats [BLOCK_SIZE][BLOCK_SIZE+1];
		
							// Load the matrices from global memory to shared memory;
		as  [ty][tx] = a [ia  + n * ty + tx];
		ats [ty][tx] = a [iat + n * ty + tx];
		
		__syncthreads();	// Synchronize to make sure the matrices are loaded
		
							// Multiply the two matrices together;
		for ( int k = 0; k < BLOCK_SIZE; k++ )
			sum += as [ty][k] * ats [tx][k];
			
							// Synchronize to make sure that the preceding
							// computation is done before loading two new
							// sub-matrices of A and B in the next iteration
		__syncthreads();
	}
	
							// Write the block sub-matrix to global memory;
							// each thread writes one element
	int ic = n * BLOCK_SIZE * by + BLOCK_SIZE * bx;
	
	c [ic + n * ty + tx] = sum;
}

int main ( int argc, char *  argv [] )
{
	int	numBytes = N * N * sizeof ( float );

					// allocate host memory
    float * a    = new float [N*N];
	float * b    = new float [N*N];
	float	phi  = M_PI / 3;
	float	cphi = cosf ( phi );
	float	sphi = sinf ( phi );
	
	for ( int row = 0; row < N; row++ )
		for ( int col = 0; col < N; col++ )
		{
			int	offs = row * N + col;
			
			a [offs] = 0.0f;
			
			if ( row == col )
				a [offs] = cphi;
			else
			if ( row & 1 )			//  
			{
				if ( col + 1 == row )
					a [offs] = -sphi;
			}
			else		//  
			{
				if ( col - 1 == row )
					a [offs] = sphi;
			}
		}
		
					// allocate device memory
    float * adev = NULL;
    float * bdev = NULL;
	
    cudaMalloc ( (void**)&adev, numBytes );
    cudaMalloc ( (void**)&bdev, numBytes );

					// set kernel launch configuration
	dim3 threads ( BLOCK_SIZE, BLOCK_SIZE );
	dim3 blocks  ( N / threads.x, N / threads.y);

					// create cuda event handles
    cudaEvent_t start, stop;
    float gpuTime = 0.0f;

    cudaEventCreate ( &start );
    cudaEventCreate ( &stop );
	
					// asynchronously issue work to the GPU (all to stream 0)
    cudaEventRecord ( start, 0 );
    cudaMemcpy      ( adev, a, numBytes, cudaMemcpyHostToDevice );
	
    matMult<<<blocks, threads>>> ( adev, N, bdev );
	
    cudaMemcpy      ( b, bdev, numBytes, cudaMemcpyDeviceToHost );
    cudaEventRecord ( stop, 0 );

	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

						// print the cpu and gpu times
    printf("time spent executing by the GPU: %.2f millseconds\n", gpuTime );

	for ( int row = 0; row < N; row++ )
		for ( int col = 0; col < N; col++ )
		{
			int	offs = row * N + col;
			
			if ( col == row )
			{
				if ( fabs ( b [offs] - 1.0f ) > EPS )
					printf ( "Error at (%d,%d)\n", row, col );
			}
			else
			{
				if ( fabs ( b [offs] ) > EPS )
					printf ( "Error at (%d,%d)\n", row, col );
			}
		}
		
					// release resources
    cudaEventDestroy ( start );
    cudaEventDestroy ( stop  );
    cudaFree         ( adev  );
    cudaFree         ( bdev  );

    delete a;
	delete b;

    return 0;
}
