//
// Perform "naive" parallel reduction
//

#include	<stdio.h>
#include	<stdlib.h>

#define BLOCK_SIZE 		256	
#define	N				(256*256)		// size of reduced array 
#define	LOG_NUM_BANKS	4

#define	CONFLICT_FREE_OFFS(i)	((i) >> LOG_NUM_BANKS)

__global__ void scan3 ( float * inData, float * outData, float * sums, int n )
{
	__shared__ float temp [2*BLOCK_SIZE+CONFLICT_FREE_OFFS(2*BLOCK_SIZE)];
	
	int	tid    = threadIdx.x;
	int	offset = 1;
	int	ai     = tid;
	int	bi     = tid + (n / 2);
	int	offsA  = CONFLICT_FREE_OFFS(ai);
	int	offsB  = CONFLICT_FREE_OFFS(bi);
	
	temp [ai + offsA] = inData [ai + 2*BLOCK_SIZE*blockIdx.x];	// load into shared memeory
	temp [bi + offsB] = inData [bi + 2*BLOCK_SIZE*blockIdx.x];
	
	for ( int d = n>>1; d > 0; d >>= 1 )
	{
		__syncthreads ();
		
		if ( tid < d )
		{
			int	ai = offset * (2 * tid + 1) - 1;
			int	bi = offset * (2 * tid + 2) - 1;
			
			ai        += CONFLICT_FREE_OFFS(ai);
			bi        += CONFLICT_FREE_OFFS(bi);
			temp [bi] += temp [ai];
		}
		
		offset <<= 1;
	}
	
	if ( tid == 0 )
	{
		int	i = n - 1 + CONFLICT_FREE_OFFS(n-1);
		
		sums [blockIdx.x] = temp [i];	// save the sum
		temp [i]          = 0;			// clear the last element
	}
		
	for ( int d = 1; d < n; d <<= 1 )
	{
		offset >>= 1;
		__syncthreads ();
		
		if ( tid < d )
		{
			int		ai = offset * (2 * tid + 1) - 1;
			int		bi = offset * (2 * tid + 2) - 1;
			float	t;
			ai        += CONFLICT_FREE_OFFS(ai);
			bi        += CONFLICT_FREE_OFFS(bi);
			t          = temp [ai];
			temp [ai]  = temp [bi];
			temp [bi] += t;
		}
	}
	
	__syncthreads ();
	
	outData [ai + 2*BLOCK_SIZE*blockIdx.x] = temp [ai + offsA];			// write results
	outData [bi + 2*BLOCK_SIZE*blockIdx.x] = temp [bi + offsB];
}

__global__ void scanDistribute ( float * data, float * sums )
{
	data [threadIdx.x+blockIdx.x*2*BLOCK_SIZE] += sums [blockIdx.x];
}

//
//perform scan on device array
//

void	scan ( float * inData, float * outData, int n )
{
	int		numBlocks = n / (2*BLOCK_SIZE);
	float * sums      = NULL;			// sums of elements for every block
	float * sums2     = NULL;			// scan of sums

	if ( numBlocks < 1 )
		numBlocks = 1;
							// allocate sums array
    cudaMalloc ( (void**)&sums,  numBlocks * sizeof ( float ) );
    cudaMalloc ( (void**)&sums2, numBlocks * sizeof ( float ) );
	
							// perform blockwise scan, 1 thread per 2 data items
	dim3 threads ( BLOCK_SIZE, 1, 1 );
	dim3 blocks  ( numBlocks, 1, 1 );
	
	scan3<<<blocks, threads>>> ( inData, outData, sums, 2*BLOCK_SIZE );

							// now scan the sums
	if ( n >= 2*BLOCK_SIZE )
		scan ( sums, sums2, numBlocks );
	else
		cudaMemcpy ( sums2, sums, numBlocks*sizeof(float), cudaMemcpyDeviceToDevice );

							// now distribute scanned sums
	threads = dim3 ( 2*BLOCK_SIZE, 1, 1 );
	blocks  = dim3 ( numBlocks - 1, 1, 1 );
	
	scanDistribute<<<blocks,threads>>> ( outData + 2*BLOCK_SIZE, sums2 + 1 );
		
    cudaFree  ( sums  );
    cudaFree  ( sums2 );
}

int main ( int argc, char *  argv [] )
{
	int	numBytes = N * sizeof ( float );
	int	n        = N;
	int	i        = 0;

					// allocate host memory
    float * a = new float [N];
	float * b = new float [N];
	
					// init with random values
	for ( i = 0; i < N; i++ )
		a [i] = 1; //(rand () & 0xFF) - 127;
		
					// allocate device memory
    float     * adev [2] = { NULL, NULL };
    cudaEvent_t start, stop;
    float 		gpuTime = 0.0f;
	
    cudaMalloc ( (void**)&adev [0], numBytes );
    cudaMalloc ( (void**)&adev [1], numBytes );

					// create cuda event handles
    cudaEventCreate ( &start );
    cudaEventCreate ( &stop );
	
					// asynchronously issue work to the GPU (all to stream 0)
    cudaEventRecord ( start, 0 );
    cudaMemcpy      ( adev [0], a, numBytes, cudaMemcpyHostToDevice );

	scan ( adev [0], adev [1], n );

    cudaMemcpy      ( b, adev [1], numBytes, cudaMemcpyDeviceToHost );
    cudaEventRecord ( stop, 0 );

	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	for ( i = 0; i < 3000; i++ )
		if ( fabs ( b [i] - i ) > 0.0001 )
			printf ( "item at %d diff %f --> (%f %d)\n", i, b [i] - i, b [i], i );
		
	printf ( "\n" );
						// print the cpu and gpu times
    printf ( "time spent executing by the GPU: %.2f milliseconds\n", gpuTime );

					// release resources
    cudaEventDestroy ( start );
    cudaEventDestroy ( stop  );
    cudaFree         ( adev [0] );
    cudaFree         ( adev [1] );

    delete a;
	delete b;

    return 0;
}
