#include	<cudpp.h>
#include	<stdio.h>
#include	<stdlib.h>

#define	N			(5*1024*1024)

typedef	unsigned int	uint;

// -----------------------------------------------------------------------------------------------
// The floatFlip and floatUnflip functions below are based on code in the web article 
// "Radix Tricks" by Michael Herf (http://www.stereopsis.com/radix.html). They are used to convert
// floating point values into sortable unsigned integers (and back).
//
// Paraphrasing Michael: Binary single-precision floating point numbers have two features that 
// keep them from being directly sortable. First, the sign bit is set when the value is negative, 
// which means that all negative numbers are bigger than positive ones. Second, the values are 
// signed-magnitude, so "more negative" floating point numbers actually look bigger to a normal 
// bitwise comparison.
// 
// "To fix our floating point numbers, we define the following rules:
//
//   1. Always flip the sign bit.
//   2. If the sign bit was set, flip the other bits too.
//
// To get back, we flip the sign bit always, and if the sign bit was not set, we flip the other 
// bits too."
//
// This is a very inexpensive operation and it is only done on the first and last steps of the
// sort.
// -----------------------------------------------------------------------------------------------


// ================================================================================================
// Flip a float for sorting
//  finds SIGN of fp number.
//  if it's 1 (negative float), it flips all bits
//  if it's 0 (positive float), it flips the sign only
// ================================================================================================
template <bool doFlip>
__device__ uint floatFlip(uint f)
{
    if (doFlip)
    {
        uint mask = -int(f >> 31) | 0x80000000;
		return f ^ mask;
    }
    else
        return f;
}

// ================================================================================================
// flip a float back (invert FloatFlip)
//  signed was flipped from above, so:
//  if sign is 1 (negative), it flips the sign bit back
//  if sign is 0 (positive), it flips all bits back
// ================================================================================================
template <bool doFlip>
__device__ uint floatUnflip(uint f)
{
    if (doFlip)
    {
        uint mask = ((f >> 31) - 1) | 0x80000000;
	    return f ^ mask;
    }
    else
        return f;
}

// ================================================================================================
// Kernel to flip all floats in an array (see floatFlip, above)
// Each thread flips four values (each 256-thread CTA flips 1024 values).
// ================================================================================================
__global__ void flipFloats ( uint * values, uint numValues)
{
    uint index = __umul24(blockDim.x*4, blockIdx.x) + threadIdx.x; 
	
    if (index < numValues) 
		values[index] = floatFlip<true>(values[index]);
		
    index += blockDim.x;
	
    if (index < numValues) 
		values[index] = floatFlip<true>(values[index]);
		
    index += blockDim.x;
	
    if (index < numValues) 
		values[index] = floatFlip<true>(values[index]);
		
    index += blockDim.x;
	
    if (index < numValues) 
		values[index] = floatFlip<true>(values[index]);
}

// ================================================================================================
// Kernel to unflip all floats in an array (see floatUnflip, above)
// Each thread unflips four values (each 256-thread CTA unflips 1024 values).
// ================================================================================================
__global__ void unflipFloats ( uint * values, uint numValues )
{
    uint index = __umul24(blockDim.x*4, blockIdx.x) + threadIdx.x; 
	
    if (index < numValues) 
		values [index] = floatUnflip<true>(values [index]);
		
    index += blockDim.x;
	
    if (index < numValues) 
		values [index] = floatUnflip<true>(values [index]);
    index += blockDim.x;
    if (index < numValues) 
		values [index] = floatUnflip<true>(values [index]);
		
    index += blockDim.x;
	
    if (index < numValues) 
		values [index] = floatUnflip<true>(values [index]);
}

void randomInit ( float * a, int n )
{
	for ( int i = 0; i < n; i++ )
		a [i] = rand () / (float) RAND_MAX;
}

int main ( int argc, char * argv [] )
{                
	CUDPPConfiguration	config;
	CUDPPAlgorithm 		algos [] = { CUDPP_SORT_RADIX, CUDPP_SORT_RADIX_GLOBAL };
	CUDPPHandle			plan;
    CUDPPResult			result      = CUDPP_SUCCESS;
	float			  * array       = new float [N];
	uint              * arrayDevIn  = NULL;
	uint              * arrayDevOut = NULL;
    cudaEvent_t 		start, stop;
    float 				gpuTime = 0.0f;
	
	randomInit ( array, N );
	
    cudaEventCreate ( &start );
    cudaEventCreate ( &stop );
    cudaEventRecord ( start, 0 );
	
	cudaMalloc ( (void **) &arrayDevIn,  N * sizeof ( int ) );
	cudaMalloc ( (void **) &arrayDevOut, N * sizeof ( int ) );
    cudaMemcpy ( arrayDevIn, array, N * sizeof ( int ), cudaMemcpyHostToDevice );
	
	flipFloats<<<dim3(N/1024), dim3(256)>>> ( arrayDevIn, N );
	
	config.datatype  = CUDPP_INT;
	config.op        = CUDPP_ADD;
	config.options   = (CUDPPOption)0;
	config.algorithm = algos [0];
            
    result = cudppPlan ( &plan, config, N, 1, 0 );

    if ( result != CUDPP_SUCCESS )
    {
        fprintf ( stderr, "Error creating plan for Sort\n" );
        return 1;
    }
                
    cudppSort  ( plan, arrayDevOut, arrayDevIn, N );

	unflipFloats<<<dim3(N/1024), dim3(256)>>> ( arrayDevOut, N );
	
    cudaMemcpy ( array, arrayDevOut, N * sizeof ( int ), cudaMemcpyDeviceToHost );

	cudppDestroyPlan ( plan );
	
	cudaFree   ( arrayDevIn  );
	cudaFree   ( arrayDevOut );
	
    cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	printf ( "Elapsed time: %.2f\n", gpuTime );
	
	for ( int i = 1; i < N; i++ )
		if ( array [i-1] > array [i] )
			printf ( "Error at pos %d\n", i );
		
	return 0;
}